package tools.descartes.dql.core.engine.interpreter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Comparator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.emf.common.util.ECollections;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.IDE;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.util.EditorHelper;
import tools.descartes.dql.core.engine.util.QueryParserHelper;
import tools.descartes.dql.lang.descartesQL.ArithmeticClause;
import tools.descartes.dql.lang.descartesQL.DescartesQL;
import tools.descartes.dql.lang.descartesQL.EntityReference;
import tools.descartes.dql.lang.descartesQL.GenerateSLAConcern;
import tools.descartes.dql.lang.descartesQL.ModelFamily;
import tools.descartes.dql.lang.descartesQL.ModelReferenceClause;
import tools.descartes.dql.lang.descartesQL.MultipleArithmeticClause;
import tools.descartes.dql.lang.descartesQL.PerformanceMetricsQuery;
import tools.descartes.dql.lang.descartesQL.UsingClause;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DiscreteSeriesResult;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Result;
import tools.descartes.dql.models.mapping.domain.SeriesResultElement;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;

public class SLAGeneratorInterpreter
		extends AbstractQueryInterpreter<GenerateSLAConcern, PerformanceMetricsQueryConnector> {

	private final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private DefaultProvider defaultProvider;
	private MultipleArithmeticClause offsets;
	private static String constraintClause;
	private static String variationClause;
	private double satisfactionLevel = 100.0;

	public SLAGeneratorInterpreter(final GenerateSLAConcern query) {
		super(query);
	}

	public SLAGeneratorInterpreter(final GenerateSLAConcern query, List<QueryProgressListener> listeners) {
		super((GenerateSLAConcern) query, listeners);
	}

	@Override
	public boolean interpretQuery() {
		GenerateSLAConcern generateSLAConcern;

		if (getQuery() instanceof GenerateSLAConcern) {
			generateSLAConcern = (GenerateSLAConcern) getQuery();
			final UsingClause usingClause = generateSLAConcern.getUsingClause();
			final ModelReferenceClause modelReference = usingClause.getModelReference();
			final ModelFamily family = modelReference.getFamily();
			final String familyIdentifier = family.getFamily();
			setModelFamily(familyIdentifier);

			generateSLA(generateSLAConcern);
			return true;
		} else {
			log.error("Query type not supported");
			return false;
		}
	}

	private void generateSLA(GenerateSLAConcern generateSLAConcern) {
		this.defaultProvider = new DefaultProvider(generateSLAConcern.getUsingClause());
		final UsingClause usingClause = generateSLAConcern.getUsingClause();
		final ModelReferenceClause modelReference = usingClause.getModelReference();
		final ModelFamily family = modelReference.getFamily();
		final String familyIdentifier = family.getFamily();

		constraintClause = "";
		if (generateSLAConcern.getConstraintClause() != null) {
			constraintClause = "\nCONSTRAINED AS '" + generateSLAConcern.getConstraintClause().getType().toString()
					+ "'\n";
		}
		variationClause = SLAInterpreter.extractVariationString(generateSLAConcern.getDofClause());

		try {
			satisfactionLevel = generateSLAConcern.getPercentage();
		} catch (NullPointerException e) {
		}

		setModelFamily(familyIdentifier);

		if (!initQueryConnector()) {
			String exception = "Could not load Connector for family " + getModelFamily() + ", aborting.";
			log.error(exception);
			throw new IllegalStateException(exception);
		}

		// Reset before work, allowing independent caching mechanisms in
		// the DQL Connector
		getQueryConnector().reset();

		// ModelLocation model = modelReference.getModel();
		// String location = model.getLocation();
		// final EntityMapping request = mappingFactory.createEntityMapping();
		// request.setQueryAsString(getQueryAsString());
		// request.setModelLocation(location);
		// TODO fix relative paths
		// if (!new File(location).exists()) {
		// location = QueryExecutionEngineImpl
		// .resolveRelativePath((new
		// File(model.getURI().path())).getParentFile().getAbsolutePath(), location);
		// usingClause.getModelReference().getModel().setLocation(location);
		// }

		StringBuffer metricsQuerySB = new StringBuffer();
		metricsQuerySB.append("SELECT\n");
		for (EntityReference er : generateSLAConcern.getForClause().getEntityReferenceClause().getEntityReferences()) {
			if (er.getType().equalsIgnoreCase("SERVICE")) {
				metricsQuerySB.append("\t" + er.getAlias().getAlias() + "." + "responseTime" + ",\n");
				// + "responseTime95thPerc" + ",\n");
				// + "responseTime99thPerc" + ",\n");
			}
		}
		metricsQuerySB.deleteCharAt(metricsQuerySB.length() - 2); // remove last comma

		String queryString = getQueryAsString();
		metricsQuerySB.append(constraintClause);
		metricsQuerySB.append(variationClause);

		String endString = "FOR" + queryString.split("FOR")[1] + ";";
		endString = endString.replace("%20", " ");
		endString = endString.replace(" @ ", "@");
		endString = endString.replace(" @\r\n", "@");
		endString = endString.replace(" ,", ",");

		metricsQuerySB.append(endString);

		String metricsQuery = metricsQuerySB.toString();
		metricsQuery = metricsQuery.replace(" ,", ",");

		log.info("-------------------\n" + "---metrics concern---\n" + metricsQuery + "\n-------------------");

		EntityMapping result = runPerformanceMetricQuery(metricsQuery);

		offsets = generateSLAConcern.getOffset();

		String generatedSLA = generateSLA(result, endString);

		String path = generateSLAConcern.getUsingClause().getModelReference().getModel().getLocation().toString();
		path = (new File(path)).getParent().toString() + File.separator + "generated_sla.dql";
		writeToFile(path, generatedSLA);
		EditorHelper.openInEditor(path);
	}

	private void openEditor(IFile file) {
		final IFile aFile = file;
		PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
			public void run() {
				IWorkbenchWindow dwindow = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
				IWorkbenchPage page = dwindow.getActivePage();
				if (page != null) {
					try {
						IDE.openEditor(page, aFile, true);
					} catch (PartInitException pie) {
						log.error("Unable to open the Editor", pie);
					}
				}
			}
		});
	}

	private String generateSLA(EntityMapping result, String endString) {
		StringBuffer slaString = new StringBuffer();
		StringBuffer sloString = new StringBuffer();

		slaString.append("MONITOR\n");
		int sloCnt = 1;
		double value = 0.0;
		for (Entity ent : result.getEntities()) {
			Result subresult = ent.getProbes().get(0).getStatTypes().get(0).getResult();
			if (subresult instanceof DiscreteSeriesResult) {
				value = 0.0;
				for (SeriesResultElement elem : ((DiscreteSeriesResult) subresult).getElements()) {
					value = Math.max(value, elem.getY().doubleValue());
				}
				value = getPercentile((DiscreteSeriesResult) subresult, satisfactionLevel);
			} else if (subresult instanceof DecimalResult) {
				value = (((DecimalResult) subresult).getValue()).doubleValue();
			}
			for (ArithmeticClause offset : offsets.getMultiClause()) {
				double offsetValue = offset.getValue().getDoubleValue();

				if (offset.getMode().equalsIgnoreCase("MULTIPLY")) {
					value = value * offsetValue;
					log.info("multiplied " + ent.getAlias() + " " + ent.getProbes().get(0).getMetricName()
							+ " threshold by " + offsetValue);
				} else if (offset.getMode().equalsIgnoreCase("ADD") | offset.getMode().equalsIgnoreCase("+")) {
					value = value + offsetValue;
					log.info("added " + offsetValue + " offset to " + ent.getAlias() + " "
							+ ent.getProbes().get(0).getMetricName() + " threshold");
				}
			}
			value = Math.floor(value * 1000d) / 1000d;
			// Math.round(value * 10000.0) / 10000.0)
			sloString.append("\t\tslo" + sloCnt++ + ": " + ent.getAlias() + "." + ent.getProbes().get(0).getMetricName()
					+ "<" + value + " SATISFACTION-LEVEL " + satisfactionLevel + " PERCENT" + ",\n");
		}
		sloString.deleteCharAt(sloString.length() - 2);

		slaString.append("\tAGREEMENTS\n\t\tsla1 CONTAINS (");
		for (int i = 1; i < sloCnt; i++) {
			slaString.append("slo" + i + ",");
		}
		slaString.deleteCharAt(slaString.length() - 1); // remove last comma
		slaString.append(")");
		slaString.append("\n\tGOALS\n");
		slaString.append(sloString);
		slaString.append(constraintClause);
		slaString.append(variationClause);
		slaString.append(endString);
		return slaString.toString();
	}

	private double getPercentile(DiscreteSeriesResult subresult, double percent) {
		EList<SeriesResultElement> list = subresult.getElements();
		EList<SeriesResultElement> list2 = EcoreUtil.copy(subresult).getElements();
		sort(list2);
		int index = (int) (list2.size() * (percent / 100.0));
		index = Math.min(index, list2.size() - 1);
		double value = list2.get(index).getY().doubleValue();
		// System.out.println(list2.get(0).getY().doubleValue() + " < " +
		// list2.get(index).getY().doubleValue() + " < " + list2.get(list2.size() -
		// 1).getY().doubleValue());

		return value;
	}

	private static void sort(EList<SeriesResultElement> list) {
		try {
			// ); list.sort(
			ECollections.sort(list, new Comparator<Object>() {
				@Override
				public int compare(Object o1, Object o2) {
					if ((((SeriesResultElement) o1).getY().doubleValue()
							- ((SeriesResultElement) o2).getY().doubleValue()) > 0) {
						return 1;
					} else if ((((SeriesResultElement) o1).getY().doubleValue()
							- ((SeriesResultElement) o2).getY().doubleValue()) < 0) {
						return -1;
					} else {
						// if ((((SeriesResultElement)o1).getX().doubleValue() -
						// ((SeriesResultElement)o2).getX().doubleValue()) > 0) {
						// return 1;
						// } else if ((((SeriesResultElement)o1).getX().doubleValue() -
						// ((SeriesResultElement)o2).getX().doubleValue()) < 0) {
						// return -1;
						// }
						if ((((SeriesResultElement) o1).getX().intValue()
								- ((SeriesResultElement) o2).getX().intValue()) > 0) {
							return 1;
						} else if ((((SeriesResultElement) o1).getX().intValue()
								- ((SeriesResultElement) o2).getX().intValue()) < 0) {
							return -1;
						}
					}
					return 1; // 0;
				}
			});
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		}
	}

	private EntityMapping runPerformanceMetricQuery(String queryString) {
		final DescartesQL queryModel = QueryParserHelper.stringToModel(queryString);
		EObject genQuery = queryModel.getQuery();

		// Reuse of Performance Metrics connector
		PerformanceMetricsQueryInterpreter interpreter = new PerformanceMetricsQueryInterpreter(
				(PerformanceMetricsQuery) genQuery);
		boolean validQuery = interpreter.interpretQuery();
		if (!validQuery) {
			log.error("query is not valid");
		}
		EntityMapping entityMapping = interpreter.getResult();
		return entityMapping;
	}

	@Override
	protected Class<PerformanceMetricsQueryConnector> getQueryConnectorClass() {
		return PerformanceMetricsQueryConnector.class;
	}

	private void writeToFile(String file, String content) {
		System.out.println("Saving SLA to " + file);
		BufferedWriter writer = null;
		try {
			writer = new BufferedWriter(new FileWriter(file));
			writer.write(content);
			System.out.println("-------------------");
			System.out.println(content);
			System.out.println("-------------------");
			System.out.println("Done");
		} catch (IOException e) {
			log.error(e);
		} finally {
			try {
				if (writer != null)
					writer.close();
			} catch (IOException e) {
				log.error(e);
			}
		}
	}

}
